/* Generated By:JavaCC: Do not edit this line. ADQLParserConstants.java */
package adql.parser;


/**
 * Token literal values and constants.
 * Generated by org.javacc.parser.OtherFilesGen#start()
 */
public interface ADQLParserConstants {

  /** End of File. */
  int EOF = 0;
  /** RegularExpression Id. */
  int SQL_RESERVED_WORD = 2;
  /** RegularExpression Id. */
  int LEFT_PAR = 3;
  /** RegularExpression Id. */
  int RIGHT_PAR = 4;
  /** RegularExpression Id. */
  int DOT = 5;
  /** RegularExpression Id. */
  int COMMA = 6;
  /** RegularExpression Id. */
  int EOQ = 7;
  /** RegularExpression Id. */
  int CONCAT = 8;
  /** RegularExpression Id. */
  int PLUS = 9;
  /** RegularExpression Id. */
  int MINUS = 10;
  /** RegularExpression Id. */
  int ASTERISK = 11;
  /** RegularExpression Id. */
  int DIVIDE = 12;
  /** RegularExpression Id. */
  int EQUAL = 13;
  /** RegularExpression Id. */
  int NOT_EQUAL = 14;
  /** RegularExpression Id. */
  int LESS_THAN = 15;
  /** RegularExpression Id. */
  int LESS_EQUAL_THAN = 16;
  /** RegularExpression Id. */
  int GREATER_THAN = 17;
  /** RegularExpression Id. */
  int GREATER_EQUAL_THAN = 18;
  /** RegularExpression Id. */
  int SELECT = 19;
  /** RegularExpression Id. */
  int QUANTIFIER = 20;
  /** RegularExpression Id. */
  int TOP = 21;
  /** RegularExpression Id. */
  int FROM = 22;
  /** RegularExpression Id. */
  int AS = 23;
  /** RegularExpression Id. */
  int NATURAL = 24;
  /** RegularExpression Id. */
  int INNER = 25;
  /** RegularExpression Id. */
  int OUTER = 26;
  /** RegularExpression Id. */
  int RIGHT = 27;
  /** RegularExpression Id. */
  int LEFT = 28;
  /** RegularExpression Id. */
  int FULL = 29;
  /** RegularExpression Id. */
  int JOIN = 30;
  /** RegularExpression Id. */
  int ON = 31;
  /** RegularExpression Id. */
  int USING = 32;
  /** RegularExpression Id. */
  int WHERE = 33;
  /** RegularExpression Id. */
  int AND = 34;
  /** RegularExpression Id. */
  int OR = 35;
  /** RegularExpression Id. */
  int NOT = 36;
  /** RegularExpression Id. */
  int IS = 37;
  /** RegularExpression Id. */
  int NULL = 38;
  /** RegularExpression Id. */
  int BETWEEN = 39;
  /** RegularExpression Id. */
  int LIKE = 40;
  /** RegularExpression Id. */
  int IN = 41;
  /** RegularExpression Id. */
  int EXISTS = 42;
  /** RegularExpression Id. */
  int BY = 43;
  /** RegularExpression Id. */
  int GROUP = 44;
  /** RegularExpression Id. */
  int HAVING = 45;
  /** RegularExpression Id. */
  int ORDER = 46;
  /** RegularExpression Id. */
  int ASC = 47;
  /** RegularExpression Id. */
  int DESC = 48;
  /** RegularExpression Id. */
  int AVG = 49;
  /** RegularExpression Id. */
  int MAX = 50;
  /** RegularExpression Id. */
  int MIN = 51;
  /** RegularExpression Id. */
  int SUM = 52;
  /** RegularExpression Id. */
  int COUNT = 53;
  /** RegularExpression Id. */
  int BOX = 54;
  /** RegularExpression Id. */
  int CENTROID = 55;
  /** RegularExpression Id. */
  int CIRCLE = 56;
  /** RegularExpression Id. */
  int POINT = 57;
  /** RegularExpression Id. */
  int POLYGON = 58;
  /** RegularExpression Id. */
  int REGION = 59;
  /** RegularExpression Id. */
  int CONTAINS = 60;
  /** RegularExpression Id. */
  int INTERSECTS = 61;
  /** RegularExpression Id. */
  int AREA = 62;
  /** RegularExpression Id. */
  int COORD1 = 63;
  /** RegularExpression Id. */
  int COORD2 = 64;
  /** RegularExpression Id. */
  int COORDSYS = 65;
  /** RegularExpression Id. */
  int DISTANCE = 66;
  /** RegularExpression Id. */
  int ABS = 67;
  /** RegularExpression Id. */
  int CEILING = 68;
  /** RegularExpression Id. */
  int DEGREES = 69;
  /** RegularExpression Id. */
  int EXP = 70;
  /** RegularExpression Id. */
  int FLOOR = 71;
  /** RegularExpression Id. */
  int LOG = 72;
  /** RegularExpression Id. */
  int LOG10 = 73;
  /** RegularExpression Id. */
  int MOD = 74;
  /** RegularExpression Id. */
  int PI = 75;
  /** RegularExpression Id. */
  int POWER = 76;
  /** RegularExpression Id. */
  int RADIANS = 77;
  /** RegularExpression Id. */
  int RAND = 78;
  /** RegularExpression Id. */
  int ROUND = 79;
  /** RegularExpression Id. */
  int SQRT = 80;
  /** RegularExpression Id. */
  int TRUNCATE = 81;
  /** RegularExpression Id. */
  int ACOS = 82;
  /** RegularExpression Id. */
  int ASIN = 83;
  /** RegularExpression Id. */
  int ATAN = 84;
  /** RegularExpression Id. */
  int ATAN2 = 85;
  /** RegularExpression Id. */
  int COS = 86;
  /** RegularExpression Id. */
  int COT = 87;
  /** RegularExpression Id. */
  int SIN = 88;
  /** RegularExpression Id. */
  int TAN = 89;
  /** RegularExpression Id. */
  int STRING_LITERAL = 93;
  /** RegularExpression Id. */
  int DELIMITED_IDENTIFIER = 96;
  /** RegularExpression Id. */
  int REGULAR_IDENTIFIER = 97;
  /** RegularExpression Id. */
  int Letter = 98;
  /** RegularExpression Id. */
  int SCIENTIFIC_NUMBER = 99;
  /** RegularExpression Id. */
  int UNSIGNED_FLOAT = 100;
  /** RegularExpression Id. */
  int UNSIGNED_INTEGER = 101;
  /** RegularExpression Id. */
  int DIGIT = 102;

  /** Lexical state. */
  int DEFAULT = 0;
  /** Lexical state. */
  int WithinString = 1;
  /** Lexical state. */
  int WithinDelimitedId = 2;

  /** Literal token values. */
  String[] tokenImage = {
    "<EOF>",
    "<token of kind 1>",
    "<SQL_RESERVED_WORD>",
    "\"(\"",
    "\")\"",
    "\".\"",
    "\",\"",
    "\";\"",
    "\"||\"",
    "\"+\"",
    "\"-\"",
    "\"*\"",
    "\"/\"",
    "\"=\"",
    "<NOT_EQUAL>",
    "\"<\"",
    "\"<=\"",
    "\">\"",
    "\">=\"",
    "\"SELECT\"",
    "<QUANTIFIER>",
    "\"TOP\"",
    "\"FROM\"",
    "\"AS\"",
    "\"NATURAL\"",
    "\"INNER\"",
    "\"OUTER\"",
    "\"RIGHT\"",
    "\"LEFT\"",
    "\"FULL\"",
    "\"JOIN\"",
    "\"ON\"",
    "\"USING\"",
    "\"WHERE\"",
    "\"AND\"",
    "\"OR\"",
    "\"NOT\"",
    "\"IS\"",
    "\"NULL\"",
    "\"BETWEEN\"",
    "\"LIKE\"",
    "\"IN\"",
    "\"EXISTS\"",
    "\"BY\"",
    "\"GROUP\"",
    "\"HAVING\"",
    "\"ORDER\"",
    "\"ASC\"",
    "\"DESC\"",
    "\"AVG\"",
    "\"MAX\"",
    "\"MIN\"",
    "\"SUM\"",
    "\"COUNT\"",
    "\"BOX\"",
    "\"CENTROID\"",
    "\"CIRCLE\"",
    "\"POINT\"",
    "\"POLYGON\"",
    "\"REGION\"",
    "\"CONTAINS\"",
    "\"INTERSECTS\"",
    "\"AREA\"",
    "\"COORD1\"",
    "\"COORD2\"",
    "\"COORDSYS\"",
    "\"DISTANCE\"",
    "\"ABS\"",
    "\"CEILING\"",
    "\"DEGREES\"",
    "\"EXP\"",
    "\"FLOOR\"",
    "\"LOG\"",
    "\"LOG10\"",
    "\"MOD\"",
    "\"PI\"",
    "\"POWER\"",
    "\"RADIANS\"",
    "\"RAND\"",
    "\"ROUND\"",
    "\"SQRT\"",
    "\"TRUNCATE\"",
    "\"ACOS\"",
    "\"ASIN\"",
    "\"ATAN\"",
    "\"ATAN2\"",
    "\"COS\"",
    "\"COT\"",
    "\"SIN\"",
    "\"TAN\"",
    "<token of kind 90>",
    "\"\\\'\"",
    "<token of kind 92>",
    "\"\\\'\"",
    "\"\\\"\"",
    "<token of kind 95>",
    "\"\\\"\"",
    "<REGULAR_IDENTIFIER>",
    "<Letter>",
    "<SCIENTIFIC_NUMBER>",
    "<UNSIGNED_FLOAT>",
    "<UNSIGNED_INTEGER>",
    "<DIGIT>",
  };

}
